/*
 * Copyright (c) 2015-2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 *  ======== Hwi.c ========
 */

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Startup.h>

#include <ti/sysbios/BIOS.h>
#include "package/internal/Hwi.xdc.h"

/*
 *  ======== Hwi_Module_startup ========
 */
Int Hwi_Module_startup(Int phase)
{

    /* must wait for these modules to initialize first */
    if (!Hwi_HwiProxy_Module_startupDone()) {
        return Startup_NOTDONE;
    }

    /* okay to proceed with initialization */

    return Startup_DONE;
}

/*
 *  ======== Hwi_disableInterrupt ========
 */
UInt Hwi_disableInterrupt(UInt intNum)
{
    return (Hwi_HwiProxy_disableInterrupt(intNum));
}

/*
 *  ======== Hwi_enableInterrupt ========
 */
UInt Hwi_enableInterrupt(UInt intNum)
{
    return (Hwi_HwiProxy_enableInterrupt(intNum));
}

/*
 *  ======== Hwi_RestoreInterrupt ========
 */
Void Hwi_restoreInterrupt(UInt intNum, UInt key)
{
    Hwi_HwiProxy_restoreInterrupt(intNum, key);
}

/*
 *  ======== Hwi_clearInterrupt ========
 */
Void Hwi_clearInterrupt(UInt intNum)
{
    Hwi_HwiProxy_clearInterrupt(intNum);
}

/*
 *  ======== Hwi_post ========
 */
Void Hwi_post(UInt intNum)
{
    Hwi_HwiProxy_post(intNum);
}

/*
 *  ======== Hwi_getStackInfo ========
 */
Bool Hwi_getStackInfo(Hwi_StackInfo *stkInfo, Bool computeStackDepth)
{
    return(Hwi_HwiProxy_getStackInfo(stkInfo, computeStackDepth));
}

/*
 *  ======== Hwi_getCoreStackInfo ========
 */
Bool Hwi_getCoreStackInfo(Hwi_StackInfo *stkInfo, Bool computeStackDepth,
    UInt coreId)
{
    return(Hwi_HwiProxy_getCoreStackInfo(stkInfo, computeStackDepth, coreId));
}

/*
 *  ======== Hwi_getTaskSP ========
 */
Char *Hwi_getTaskSP()
{
    return (Hwi_HwiProxy_getTaskSP());
}

/*
 *  ======== switchFromBootStack ========
 */
Void Hwi_switchFromBootStack()
{
    Hwi_HwiProxy_switchFromBootStack();
}


/*
 *  ======== Hwi_Instance_init ========
 */
Int Hwi_Instance_init(Hwi_Object *hwi, Int intNum, Hwi_FuncPtr fxn, const Hwi_Params *params, Error_Block *eb)
{
    hwi->pi = Hwi_HwiProxy_create(intNum, fxn, (Hwi_HwiProxy_Params *)params, eb);
    
    if (hwi->pi == NULL) {
        return (1);
    }
    return (0);
}

/*
 *  ======== Hwi_Instance_finalize ========
 */
Void Hwi_Instance_finalize(Hwi_Object *hwi, Int status)
{
    /*
     * Only call HwiProxy_delete() on normal Hwi_delete() calls.
     * If Hwi_HwiProxy_create() failed, Hwi_HwiProxy_Instance_finalize
     * was already called automatically.
     */
    if (status == 0) {
        Hwi_HwiProxy_delete(&hwi->pi);
    }
}

/*
 *  ======== Hwi_getfunc ========
 */
Hwi_FuncPtr Hwi_getFunc(Hwi_Object *hwi, UArg *arg)
{
    return (Hwi_HwiProxy_getFunc(hwi->pi, arg));
}

/*
 *  ======== Hwi_setfunc ========
 */
Void Hwi_setFunc(Hwi_Object *hwi, Hwi_FuncPtr fxn, UArg arg)
{
    Hwi_HwiProxy_setFunc(hwi->pi, fxn, arg);
}

/*
 *  ======== Hwi_getIrp ========
 */
Hwi_Irp Hwi_getIrp(Hwi_Object *hwi)
{
    return (Hwi_HwiProxy_getIrp(hwi->pi));
}

/*
 *  ======== Hwi_getHookContext ========
 */
Ptr Hwi_getHookContext(Hwi_Object *hwi, Int id)
{
    return (Hwi_HwiProxy_getHookContext((Hwi_HwiProxy_Handle)hwi, id));
}

/*
 *  ======== Hwi_setHookContext ========
 */
Void Hwi_setHookContext(Hwi_Object *hwi, Int id, Ptr hookContext)
{
    Hwi_HwiProxy_setHookContext((Hwi_HwiProxy_Handle)hwi, id, hookContext);
}

/*
 * Below APIs provided for backward compatability.
 */

#ifdef __ti__
#pragma CODE_SECTION(ti_sysbios_hal_Hwi_disable__E, ".text:ti_sysbios_hal_Hwi_disable__E")
#pragma CODE_SECTION(ti_sysbios_hal_Hwi_enable__E, ".text:ti_sysbios_hal_Hwi_enable__E")
#pragma CODE_SECTION(ti_sysbios_hal_Hwi_restore__E, ".text:ti_sysbios_hal_Hwi_restore__E")
#pragma FUNC_EXT_CALLED(ti_sysbios_hal_Hwi_disable__E);
#pragma FUNC_EXT_CALLED(ti_sysbios_hal_Hwi_enable__E);
#pragma FUNC_EXT_CALLED(ti_sysbios_hal_Hwi_restore__E);
#endif

/*
 *  ======== Hwi_disable ========
 */
UInt ti_sysbios_hal_Hwi_disable__E()
{
    return Hwi_HwiProxy_disable();
}

/*
 *  ======== Hwi_enable ========
 */
UInt ti_sysbios_hal_Hwi_enable__E()
{
    return Hwi_HwiProxy_enable();
}

/*
 *  ======== Hwi_restore ========
 */
Void ti_sysbios_hal_Hwi_restore__E(UInt key)
{
    Hwi_HwiProxy_restore(key);
}


